/*-
 * Copyright (c) 1998 John Galbraith
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Mon Sep 21 23:07:05 MST 1998
 *
 * Other Contributors:
 *    Mike Smith <mike@smith.net.au>  
 *              -- lots of good FreeBSD advice
 *    Thomas Gellekum <tg@ihf.rwth-aachen.de> 
 *              -- PnP support
 *    Steven G. Kargl <kargl@troutmask.apl.washington.edu> 
 *              -- FreeBSD 3.0 support
 */

#include "gp.h"
#include "pnp.h"

/* this EXTRA_GP stuff and NGPTOT is to accomodate PNP devices */
#ifndef EXTRA_GP
#if NPNP > 0
#define EXTRA_GP 2
#else
#define EXTRA_GP 0
#endif
#endif

#define NGPTOT (NGP + EXTRA_GP)

/* if we have any GP's configured */
#if NGP > 0

/* #define GPIB_DEBUG */
#define GPIB_PARANOID

#include <sys/ioccom.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/proc.h>
#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/buf.h>
#include <sys/types.h>
#include <sys/cdefs.h>
#include <sys/uio.h>
#include <sys/malloc.h>
#include </sys/i386/isa/isa.h>
#include </sys/i386/isa/isa_device.h>
#include </sys/i386/isa/icu.h>

#ifdef ACTUALLY_LKM_NOT_KERNEL
static void gpintr(int unit);
#include "gpib.h"
#include "gpibreg.h"
#define GP_MODULE
#else
#include "i386/isa/gpib.h"
#include "i386/isa/gpibreg.h"
#endif

#if NPNP > 0
#include <i386/isa/pnp.h>
#endif

#define GP_CDEV_MAJOR 44
#define	GP_PRI		(PZERO+8)|PCATCH

/* keep this a power of two, or bad things will happen with uiomove().
   Also, keep >= 32 because that is the size of the FIFO's.  Finally,
   bad things happen if this is > 32K, because that is the size of the 
   TURBO counters. */
#define GPIB_MAX_TRANSFER 8*1024

/* use the bottom five bits of minor # for GPIB address, bit 6 
   for status ports, and the rest for unit */
#define GPADDR_FROM_DEV(dev) (minor(dev)&0x3F)
#define UNIT_FROM_DEV(dev) ((minor(dev)&0xFC0)>>6)

/* special status device minor #'s */
#define GPIB_NOCONTROLLER_PORT 0x1F
#define GPIB_STTY_MSG  0x20  /* print stty like info */
#define GPIB_POLL_PORT 0x21  /* serial poll all 31 addresses */
#define GPIB_DATA_PORT 0x22  /* read data port, return unsigned char */
#define GPIB_DATA_MSG  0x23: /* read data port, return formatted message */
#define GPIB_STATUS_PORT 0x24 /* read status lines, return unsigned char */
#define GPIB_STATUS_MSG 0x25 /* read status lines, return formatted message */

/* use unit specific identifiers for tsleep().  Sleeping processes can 
 be polling, waiting for some other process to give up the bus, or
 waiting for an interrupt generated by GPIB activity */
#define GP_POLL(unit) ((caddr_t)&gp[(unit)].poll_timeout)
#define GP_WAIT(unit) ((caddr_t)&gp[(unit)].wait_timeout)
#define GP_INT(unit) ((caddr_t)&gp[(unit)].irq_timeout)

/* Supposedly, there are all these different variants of chips to be
   found in National GPIB products.  I haven't actually seen all of
   these to check and debug this driver.  */
typedef enum {NAT4882A, NAT4882B, TNT4882, TURBO} ChipType;

/* the 7210 allows for different T1 delays.  Typical GPIB instruments
   (including National GPIB boards) use 75160 and 75162 chips for
   driving the GPIB.  In this case, T1_FAST is specified.  If, in
   addition to that requirement, the total cable length is less than
   15 meters or 1 meter times the number of equivalent device loads in 
   the system, and all devices on the GPIB are powered on, then
   T1_ULTRAFAST can be used.  Otherwise, use T1_SLOW */
typedef enum {T1_SLOW, T1_FAST, T1_ULTRAFAST} DelayType;

/* FreeBSD entrances */
static int gp_probe(struct isa_device *);
static int gp_attach(struct isa_device *);
static int gp_open(dev_t, int, int, struct proc *);
static int gp_close(dev_t, int, int, struct proc *);
static int gp_write(dev_t, struct uio *, int);
static int gp_read(dev_t, struct uio *, int);
static int gp_ioctl(dev_t dev, int cmd, caddr_t data, int flag,
                    struct proc *p);

/* these initialization routines take iobase because the gp[] 
   structure has not yet been initialized when called. */
static int gp_atgpib_probe(int iobase);
static void gp_atgpib_attach(int iobase);

/* write multiline command to GPIB */
static int gp_controller_cmd(int unit, unsigned char cmd);
static int gp_IFC(int unit); /* Interface clear command */
static int gp_GET(int unit); /* group execute trigger */
static int gp_DCL(int unit); /* device clear */
static int gp_SDC(int unit, int address);  /* selected device clear */
static int gp_LLO(int unit, int address);  /* local lockout */
static int gp_GTL(int unit, int address);  /* goto local */
static int gp_UNT(int unit); /* untalk */
static int gp_UNL(int unit); /* unlisten */
static int gp_spoll(int unit, int gpaddr, unsigned char* result);

/* manipulate GPIB controller status.  These commands don't check if
   the device IS a controller, though.  Check the gp[unit].controller 
   bit to see if the user opened the device as a controller or not */
static int gp_controller_active(int unit);
static int gp_controller_standby(int unit);
static int gp_controller_idle(int unit);

/* query and set CIC status */
static int gp_controller_in_charge(int unit);
static int gp_become_controller_in_charge(int unit);

/* addressing commands.  These do _NOT_ check to see if the device is
   opened as a controller or not. */
static int gp_listen(int unit, int gpaddr);
static int gp_talk(int unit, int gpaddr);

/*  gp_read() and gp_write() call the variants below by
   checking how the driver is configured. */
static int gp_dma_send(int unit, struct uio* uio);
static int gp_programmed_send(int unit, struct uio* uio);
static int gp_dma_read(int unit, struct uio* uio);
static int gp_programmed_read(int unit, struct uio* uio);

/* tsleep() driver and wait for an interrupt specified by the IMR3 mask.
   Interrupts from IMR0, IMR1, IMR2 have to be enabled before this call,
   IMR3 should have TCLINT set, at least.  IMR3 itself will be set in 
   this routine. */
static int gp_wait(int unit, int imr3_mask, const char* tmsg, 
                   const char* dbgmsg);

/* because there are multiple devices on the bus, and different
   processes can open each device separately, there is a possibilty
   that one process will try to access the bus while another is
   already on it.  In this case, the second process has to wait */
static int gp_wait_for_bus(int unit, int gpaddr);
static void gp_release_bus(int unit, int gpaddr);

/* manipulate the Turbo488 FIFO's by reading and writing the cnt registers */
static int gp_read_count(int unit);
static void gp_write_count(int unit, int count);
/* don't use these.  They crash some cards hard... (older ones without
   32 bit counters?) */
static int gp_read_count32(int unit);
static void gp_write_count32(int unit, int count);

/* routines to print stuff about the gpib ports, accessed by special
   minor devices */
static void gp_print_stty(int unit, char* result);
static void gp_print_status(int unit, char* result);
static void gp_print_spoll(int unit, char* result);

extern int DELAY(int);

typedef struct 
{
    int iobase;
    int irq;
    int drq;
    short drq_aquired;
    short attached;
    unsigned int open;    /* 32 bits correspond to open devices */
    unsigned int asleep;  /* 32 bits correspond to sleeping devices */
    int active;           /* address of device currently doing I/O */
    int talker;           /* address of current talker */
    int listener;         /* address of current listener */
    short controller;     /* device is controller */
    unsigned char gpib_addr;
    unsigned char eos;
    short eos_enabled;
    short eoi_enabled;
    ChipType type;
    DelayType T1_delay;
    unsigned int poll_timeout;
    unsigned int irq_timeout;
    unsigned int wait_timeout;
        /* The maximum transfer size seems to be complicated.  The TNT 
           CNT registers are either 16 or 32 bits, depending on the
           card.  However, 32 bits is a lot of memory for this driver
           to buffer all at once, so you would like to uiomove smaller 
           chunks at a time.  This works using programmed I/O, but the 
           GPIB card doesn't interrupt until the end of the whole
           transfer, so the DMA routine never knows when to stop and
           get the next uio block. Therefore, only the 16 bit counter
           is used right now, the whole transfer is buffered in the
           kernel at once, and the absolute maximum transfer size is
           64K.  Maybe this can be made 2**32 instead of 2**16
           somehow? */
    unsigned long maxtransfer;
    unsigned char* buf; /* malloc() buffer for data storage in kernel */
    short spollmessage; /* where we are in the status message */
    short sttymessage, statusmessage;
} gp_info;

gp_info gp[NGPTOT];

/* GPIB errors */
static int gp_error;  /* this should be passed to the user process */
#define GPIB_NOERROR  0
#define GPIB_NOLISTEN 1
#define GPIB_TIMEOUT  2
#define GPIB_OVERFLOW 3
#define GPIB_ABORT    4
#define GPIB_UNKNOWN  100

/* GPIB multiline commands */
#define GPIB_GTL 0x01
#define GPIB_SDC 0x04
#define GPIB_PPC 0x05
#define GPIB_GET 0x08
#define GPIB_TCT 0x09
#define GPIB_LLO 0x11
#define GPIB_DCL 0x14
#define GPIB_PPU 0x15
#define GPIB_SPE 0x18
#define GPIB_SPD 0x19
#define GPIB_MLA 0x20
#define GPIB_UNL 0x3F
#define GPIB_MTA 0x40
#define GPIB_UNT 0x5F

#ifdef GPIB_DEBUG
#define dbg_print uprintf

/* these routines basically wrap inb(), outb(), inw(), and outw().  They
   just happen to print out the operation as well. */
static int gp_read_register(int unit, int reg);
static int gp_read_register_w(int unit, int reg);
#define gp_read_register_quiet(unit, reg) inb(gp[(unit)].iobase + (reg))
static void gp_write_register(int unit, int reg, int val, const char* msg);
static void gp_write_register_quiet(int unit, int reg, int val);
static void gp_write_register_w(int unit, int reg, int val, const char* msg);

/* these are purely for debugging purposes.  They all get #defined away
   or to basic inb()'s under normal use. */
static void gp_print_bus(int unit);
static int gp_read_isr0(int unit);
static int gp_read_isr1(int unit);
static int gp_read_isr2(int unit);
static int gp_read_isr3(int unit);
static int gp_read_adsr(int unit);
static int gp_read_sts1(int unit);
static void gp_print_controller_state(int unit);

#else  /* no debugging, minimize the damage of all the debugging calls */
#define dbg_print
#define gp_read_register(unit, reg) inb(gp[(unit)].iobase + (reg))
#define gp_read_register_w(unit, reg) inw(gp[(unit)].iobase + (reg))
#define gp_read_register_quiet(unit, reg) inb(gp[(unit)].iobase + (reg))
#define gp_write_register(unit, reg, val, msg) outb(gp[(unit)].iobase + (reg), (val))
#define gp_write_register_quiet(unit, reg, val) outb(gp[(unit)].iobase + (reg), (val));
#define gp_write_register_w(unit, reg, val, msg) outw(gp[(unit)].iobase + (reg), (val))

#define gp_print_bus(unit)
#define gp_read_isr0(unit) gp_read_register(unit, ISR0)
#define gp_read_isr1(unit) gp_read_register(unit, ISR1)
#define gp_read_isr2(unit) gp_read_register(unit, ISR2)
#define gp_read_isr3(unit) gp_read_register(unit, ISR3)
#define gp_read_adsr(unit) gp_read_register(unit, ADSR)

#define gp_read_sts1(unit) gp_read_register(unit, STS1)
#define gp_print_controller_state(unit)

/* these are bogus pointers, but they are always #defined out anyway */
static char* gpib_commands;
static char* gpib_read_registers;
static char* gpib_write_registers;
#endif /* GPIB_DEBUG */

struct isa_driver gpdriver = {gp_probe, gp_attach, "gp"};

/* device switch entry table */
#if __FreeBSD__ >= 3
#include <sys/poll.h>
#define d_select_t d_poll_t
#define noselect seltrue
#endif

struct cdevsw gp_cdevsw = {
    (d_open_t *)      gp_open, /* open */
    (d_close_t *)     gp_close, /* close */
    (d_read_t *)      gp_read, /* read */
    (d_write_t *)     gp_write, /* write */
    (d_ioctl_t *)     gp_ioctl, /* ioctl */
    (d_stop_t *)      nullstop,  /* stop */
    (d_reset_t *)     nullreset, /* reset (not used) */
    (d_devtotty_t *)  nodevtotty,/* ttys */
    (d_select_t *)    noselect,  /* select */
    (d_mmap_t *)      nommap,    /* mmap */
    (d_strategy_t *)  nostrategy,/* strategy */
    (char*)           "gp",     /* name */
    NULL,                        /* struct bdevsw* */
    -1 /*     GP_CDEV_MAJOR   */             /* major number of driver */
};

int gp_probe(struct isa_device *device)
{
    int unit = device->id_unit;
    dbg_print("gp: probing gp[%d] at 0x%x, irq %d, dma %d\n",
	      device->id_unit, device->id_iobase, ffs(device->id_irq)-1,
	      device->id_drq);

    gp[unit].type = 0;
    gp[unit].iobase = device->id_iobase;

    gp_write_register(unit, AUXMR, 0x50, "page in CSR");
    switch(gp_read_register(device->id_unit, 0x06)&0xF0)
    {
        case 0x10:
            if(gp_atgpib_probe(unit))
            {
                dbg_print("gp[%d]: found a AT-GPIB chip NAT4882A.\n", unit);
                gp[unit].type = NAT4882A;
            }
            else
            {
                dbg_print("gp[%d]: bad probe.\n", unit);
                return 0;
            }
            break;
        case 0x20:
            if(gp_atgpib_probe(unit))
            {
                dbg_print("gp[%d]: found a AT-GPIB chip NAT4882B.\n", unit);
                gp[unit].type = NAT4882B;
            }
            else
            {
                dbg_print("gp[%d]: bad probe.\n", unit);
                return 0;
            }
            break;
        case 0x30:
            if(gp_atgpib_probe(unit))
            {
                dbg_print("gp[%d]: found a AT-GPIB/TNT.\n", unit);
                gp[unit].type = TNT4882;
            }
            else
            {
                dbg_print("gp[%d]: bad probe.\n", unit);
                return 0;
            }
            break;
        case 0x00:
            if(gp_atgpib_probe(unit))
            {
                dbg_print("gp[%d]: found a AT-GPIB Turbo, NEC7120.\n", unit);
                gp[unit].type = TURBO;
            }
            else
            {
                dbg_print("gp[%d]: bad probe.\n", unit);
                return 0;
            }
            break;
        default:
            dbg_print("gp[%d]: unknown probe.\n", unit);
            return 0;
    }

    return 1;
}

int gp_attach(struct isa_device *device)
{
    int unit = device->id_unit;

    gp[unit].attached = 0;
    gp[unit].gpib_addr = unit; /* this can be changed with ioctl()*/
    gp[unit].iobase = device->id_iobase;
    gp[unit].open = 0;
    gp[unit].asleep = 0;
    gp[unit].active = -1;
    gp[unit].drq = device->id_drq;
    gp[unit].drq_aquired = 0;
    gp[unit].irq = device->id_irq;
    gp[unit].poll_timeout = hz/10;
    gp[unit].irq_timeout = 0; /* no timeout here */
    gp[unit].wait_timeout = 5*hz;
    gp[unit].eos = 0x0A;
    gp[unit].eoi_enabled = 1;
    gp[unit].eos_enabled = 1;
    gp[unit].buf = NULL;
    gp[unit].spollmessage = 0;
    gp[unit].statusmessage = 0;
    gp[unit].sttymessage = 0;
    /* this malloc never gets released, because there isn't a detach() */
    gp[unit].buf = contigmalloc(GPIB_MAX_TRANSFER,  /* size */
				M_DEVBUF,	/* allocation class */
				M_NOWAIT,	/* don't sleep */
				0,		/* lowest physical address */
				(16*1024*1024)-1, /* below 16M */
				4,		/* 32-bit alignment */
				0x10000);	/* don't cross 64k boundary */

    /* is this a 3.0 thing? */
    /* kernel_map); */  /* allocate in the kernel map */

    if(gp[unit].buf == NULL)
    {
        dbg_print("gp[%d]: cannot get buffer memory\n", unit);
        return 0;
    }

        /* maxtransfer should be 1<<31 if the 32 bit counters are
           used - use gp_read_count32() and gp_write_count32().  This
           hangs one of my cards, so I don't use them right now.  The
	   -1 term is because 2's compliment only goes that far on the 
	   positive side. */
    gp[unit].maxtransfer = 1<<15 - 1;
    gp[unit].T1_delay = T1_ULTRAFAST;

    switch(gp[unit].type)
    {
        case NAT4882A:
        case NAT4882B:
        case TNT4882:
            gp_atgpib_attach(unit);
            printf("gp[%d]: type AT-GPIB/TNT", unit);
            break;
        case TURBO:
            gp_atgpib_attach(unit);
            printf("gp[%d]: type AT-GPIB, nec7210", unit);
            break;
    }

    printf(" at 0x%x", gp[unit].iobase);
    if(gp[unit].irq)
    {
        printf(", IRQ %d", ffs(gp[unit].irq)-1);
        if(gp[unit].drq)
            printf(", DRQ %d", gp[unit].drq);
    }
    else
        printf(", using polled I/O");
    printf(".\n");

    gp[unit].attached = 1;
    gp_controller_idle(unit);
    return 1;
}

int gp_open(dev_t dev, int flag, int mode, struct proc *p)
{
    int unit = UNIT_FROM_DEV(dev);
    int gpaddr = GPADDR_FROM_DEV(dev);

    dbg_print("gp[%d]: opening device %d\n", unit, gpaddr);

    if(unit >= NGPTOT)
    {
        dbg_print("gp[%d]: no hardware configured\n", unit);
        return ENXIO;
    }

    if(!gp[unit].attached)
    {
        dbg_print("gp[%d]: not attached\n", unit);
	return ENXIO;
    }

    if(gpaddr == GPIB_NOCONTROLLER_PORT)
    {
        if(gp[unit].open)
        {
	    dbg_print("gp[%d]: already opened\n", unit);
	    return EBUSY;
	}
	gp[unit].controller = 0;
    }
    else
        gp[unit].controller = 1;

    if(gp[unit].open & (1<<gpaddr))
    {
        dbg_print("gp[%d]: device %d already open\n", unit, gpaddr);
        return EBUSY;
    }

    if(gp[unit].open == 0)  /* this is first open on unit */
    {
        gp[unit].asleep = 0;
        gp[unit].active = -1;

        switch(gp[unit].T1_delay)
	{
            case T1_SLOW:
                gp_write_register(unit, AUXMR, 0xE0, "AUXRI T1 delay 2000ns");
                gp_write_register(unit, AUXMR, 0xA0, "AUXRB, clear TRI");
                break;
            case T1_FAST:
                gp_write_register(unit, AUXMR, 0xE0, "AUXRI T1 delay 500ns");
                gp_write_register(unit, AUXMR, 0xA4, "AUXRB, set TRI");
                break;
	     case T1_ULTRAFAST:
	        gp_write_register(unit, AUXMR, 0xE8, "AUXRI T1 delay 350ns");
	        gp_write_register(unit, AUXMR, 0xA4, "AUXRB, set TRI");
                break;
  	     default:
	        dbg_print("gp[%d]: invalid T1 delay\n", unit);
	  
        }

        gp_write_register(unit, AUXMR, 0x49, "AUXRG, NTNL and CHES");
        gp_write_register(unit, AUXMR, AUX_PON, "pon");
        
            /* become controller in charge */
        if(gp[unit].controller)
        {
            gp_IFC(unit);
            gp_controller_active(unit);
            gp_UNL(unit);
            gp_UNT(unit);
        }

            /* initialize DMA. */
        if(gp[unit].drq > 0)
        {
	    if(!isa_dma_acquire(gp[unit].drq))
            {
                isa_dmainit(gp[unit].drq, GPIB_MAX_TRANSFER);
                gp[unit].drq_aquired = 1;
            }
            else
	    {
                printf("gp[%d]: cannot init dma.\n", unit);
		return EBUSY;
	    }
        }
    }
    gp[unit].open |= 1<<gpaddr;
    return 0;
}

int gp_close(dev_t dev, int flag, int mode, struct proc *p)
{
    int pri;
    int unit = UNIT_FROM_DEV(dev);
    int gpaddr = GPADDR_FROM_DEV(dev);

    dbg_print("gp[%d]: closing device %d\n", unit, gpaddr);
    if(!(gp[unit].open & (1<<gpaddr)))
    {
        dbg_print("gp[%d]: device %d not open\n", unit, gpaddr);
        return -1;
    }
    

    if(gp[unit].asleep & (1<<gpaddr))
    {
        dbg_print("gp[%d]: notice - device %d was busy\n", unit, gpaddr);
        gp[unit].asleep &= ~(1<<gpaddr);
        if(gp[unit].active == gpaddr)
            gp_release_bus(unit, gpaddr);
    }

    gp[unit].open &= ~(1<<gpaddr);
    if(gp[unit].controller && (gpaddr < 31))
        gp_GTL(unit, gpaddr);

    if(gp[unit].open == 0)  /* last close on device */
    {
            /* unaddress all devices */
        if(gp[unit].controller)
        {
            gp_UNT(unit);
            gp_UNL(unit);
            gp_write_register(unit, AUXMR, AUX_CREN, "clear REN");
            gp_controller_idle(unit);
        }
        if(gp[unit].drq_aquired)
            isa_dma_release(gp[unit].drq);
    }

    return 0;
}

int gp_read(dev_t dev, struct uio *uio, int flat)
{
    int success, i;
    int unit = UNIT_FROM_DEV(dev);
    int gpaddr = GPADDR_FROM_DEV(dev);
    
    dbg_print("gp[%d]: reading from device %d\n", unit, gpaddr);

    if(uio->uio_resid < 1)
        return EINVAL;
    
    /* it doesn't make sense to read from yourself */
    if(gpaddr == gp[unit].gpib_addr)
        return EINVAL;

    if(gpaddr & 0x20) /* special message devices */
    {
        char result[4096];
	switch(gpaddr)
	{
            case GPIB_STTY_MSG:
	        gp_print_stty(unit, result);
		break;
            case GPIB_STATUS_MSG:
	        gp_print_status(unit, result);
		break;
  	    case GPIB_POLL_PORT:
	        gp_print_spoll(unit, result);
		break;
            default:
	        return 0;
	}
	uiomove(result, strlen(result), uio);
	return 0;
    }

    if(!gp_wait_for_bus(unit, gpaddr))
        return EIO;

    gp_write_register(unit, CMDR, CMDR_RESET_FIFO, "reset FIFO");
    success = 0;
    if(gp[unit].controller) /* perform addressing, get out of the way */
    {
        gp_read_adsr(unit);
	gp_print_controller_state(unit);
        if(gp_listen(unit, gp[unit].gpib_addr)) goto done;
        if(gp_talk(unit, gpaddr)) goto done;
	if(gp_controller_standby(unit)) goto done;
    }
    else /* wait for my listen address */
    {
        int s = spltty();
        if(!(gp_read_adsr(unit) & 0x04))
        {
            gp_write_register(unit, IMR2, ISR2_ADSC, "enable ADSC interrupt");
	    (void)gp_read_isr2(unit);  /* clear ISR2 */
            if(!gp_wait(unit, ISR3_TCLINT, "addr", "MLA")) goto done;
        }
	else
	    dbg_print("gp[%d]: already have MLA\n", unit);
        splx(s);
    }
#ifdef GPIB_PARANOID
    if(!(gp_read_adsr(unit) & 0x04))
    {
	dbg_print("gp[%d]: addressing error\n", unit);
	goto done;
    }
#endif
        /* do the transfer */
    if(gp[unit].drq_aquired)
        success = gp_dma_read(unit, uio);
    else
        success = gp_programmed_read(unit, uio);

 done:
    if(gp[unit].controller)
    {
	gp_controller_active(unit);
/* 	gp_UNL(unit); */
/* 	gp_UNT(unit); */
    }

    gp_release_bus(unit, gpaddr);
    if(!success)
        return EIO;
    else
        return 0;
}

int gp_write(dev_t dev, struct uio *uio, int ioflag)
{
    int count, success, tsize;
    int unit = UNIT_FROM_DEV(dev);
    int gpaddr = GPADDR_FROM_DEV(dev);
    dbg_print("gp[%d]: writing to device %d\n", unit, gpaddr);

        /* special status device */
    if(gpaddr == GPIB_STTY_MSG)
        return EPERM;

/*     if((uio->uio_resid < 1) || (uio->uio_resid > gp[unit].maxtransfer)) */
/*         return EINVAL; */

    /* it doesn't make sense to write to yourself */
    if(gpaddr == gp[unit].gpib_addr)
        return EINVAL;

    if(!gp_wait_for_bus(unit, gpaddr))
        return EIO;

    gp_write_register(unit, CMDR, CMDR_RESET_FIFO, "reset FIFO");
    success = 0;
    if(gp[unit].controller) /* perform addressing, get out of the way */
    {
        if(gp_listen(unit, gpaddr)) goto done;
        if(gp_talk(unit, gp[unit].gpib_addr)) goto done;
	if(gp_controller_standby(unit)) goto done;
    }
    else  /* wait for my talk address */
    {
        int s = spltty();
        if(!(gp_read_adsr(unit) & 0x02))
        {
            gp_write_register(unit, IMR2, ISR2_ADSC, "enable ADSC interrupt");
	    (void)gp_read_isr2(unit);  /* clear ISR2 */
            if(!gp_wait(unit, ISR3_TCLINT, "addr", "MTA")) goto done;
        }
	else
	    dbg_print("gp[%d]: already have MTA\n", unit);
        splx(s);
    }
#ifdef GPIB_PARANOID
    if(!(gp_read_adsr(unit) & 0x02))
    {
	dbg_print("gp[%d]: addressing error\n", unit);
	goto done;
    }
#endif

    if(gp[unit].drq_aquired > 0)
        success = gp_dma_send(unit, uio);
    else
        success = gp_programmed_send(unit, uio);

 done:
        /* reenable controller */
    if(gp[unit].controller)
    {
	gp_controller_active(unit);
	gp_UNL(unit);
	gp_UNT(unit);
    }

    gp_release_bus(unit, gpaddr);
    if(!success)
        return EIO;
    else
        return 0;
}

/* return 1 on sucess */
int gp_wait_for_bus(int unit, int gpaddr)
{
    if(gp[unit].active >= 0)
    {
        dbg_print("gp[%d]: device %d waiting for bus (%d has it now)\n", 
                  unit, gpaddr, gp[unit].active);
        gp[unit].asleep |= 1 << gpaddr;
        switch(tsleep(GP_WAIT(unit), GP_PRI, "wait", gp[unit].wait_timeout))
        {
            case 0:  /* normal wakeup */
                break;
            case EWOULDBLOCK:  /* timeout */
            case EINTR:
            case ERESTART:
                dbg_print("gp[%d]: device %d never got bus\n", unit, gpaddr);
                gp[unit].asleep &= ~(1<<gpaddr);
                gp_error = GPIB_TIMEOUT;
                return 0;
                break;
        }
        dbg_print("gp[%d]: device %d proceeding with read\n", unit, gpaddr);
    }
    gp[unit].asleep &= ~(1<<gpaddr);
    gp[unit].active = gpaddr;
    return 1;
}

void gp_release_bus(int unit, int gpaddr)
{
        /* gpaddr should always equal gp[unit].active here, and device 
           gpaddr shouldn't be sleeping */
    gp[unit].active = -1;
    if(gp[unit].asleep)
        wakeup_one(GP_WAIT(unit));
}

/* return 1 on sucess */
int gp_wait(int unit, int isr3_mask, const char* tmsg, const char* dbgmsg)
{
    int success, s, time;

    s = spltty();

    /* poll a little while, just in case it is almost here */
    time = 5;
    while((time-- > 0) && !(gp_read_register_quiet(unit, ISR3) & isr3_mask));
    if(time > 0)
    {
        dbg_print("gp[%d]: have %s\n", unit, dbgmsg);
        splx(s);
        return 1;
    }

    dbg_print("gp[%d]: waiting for %s ... ", unit, dbgmsg);
    if(gp[unit].irq > 0)
    {
        gp_write_register_quiet(unit, IMR3, isr3_mask);
        switch(tsleep(GP_INT(unit), GP_PRI, (char*)tmsg, 
                      gp[unit].irq_timeout))
        {
            case 0:  /* normal interrupt wakeup */
                dbg_print("got it.\n");
                success = 1;
                gp_error = GPIB_NOERROR;
                break;
            case EWOULDBLOCK:  /* UNIX timeout (not TNT4882 timer) */
                dbg_print("kernel timeout: (0x%x:0x%x)\n", isr3_mask, 
                          gp_read_register_quiet(unit, ISR3));
                success = 0;
                gp_error = GPIB_TIMEOUT;
                break;
            case EINTR: /* interrupted system call */
                dbg_print("abort");
                success = 0;
                gp_error = GPIB_ABORT;
                break;
            case ERESTART:  /* what should happen here? */
                dbg_print("restart");
                success = 0;
                gp_error = GPIB_ABORT;
                break;
            default:
                dbg_print("internal error");
                success = 0;
                gp_error = GPIB_UNKNOWN;
        }
	gp_write_register_quiet(unit, IMR3, 0x00);
    }
    else  /* not using interupts, do poll */
    {
        int time;

        time = gp[unit].poll_timeout;
        while(!(gp_read_register_quiet(unit, ISR3) & isr3_mask) && (time > 0))
        {
  	    tsleep(GP_POLL(unit), GP_PRI, "poll", 1);
	    time--;
	}
        if(time == 0)
        {
            dbg_print("timeout\n");
            success = 0;
            gp_error = GPIB_TIMEOUT;
        }
        else
        {
            dbg_print("got it\n");
            success = 1;
        }
    }
    splx(s);
    return success;
}

int gp_ioctl(dev_t dev, int cmd, caddr_t arg, int flag, struct proc *p)
{
    struct gpibconfig* config = (struct gpibconfig *)arg;
    unsigned char* spoll_result = (unsigned char*)arg;
    int unit = UNIT_FROM_DEV(dev);
    int gpaddr = GPADDR_FROM_DEV(dev);
    int error = 0;

        /* these ioctl's don't require the bus or controller status */
    if((gpaddr == GPIB_STTY_MSG) || (gpaddr == GPIB_NOCONTROLLER_PORT))
        switch(cmd)
        {
            case GPIB_GET_CONFIG:
                config->address = gp[unit].gpib_addr;
                config->poll_timeout = gp[unit].poll_timeout;
                config->irq_timeout = gp[unit].irq_timeout;
                config->wait_timeout = gp[unit].wait_timeout;
                config->eos = gp[unit].eos;
		config->eos_enabled = gp[unit].eos_enabled;
                config->eoi_enabled = gp[unit].eoi_enabled;
                return(0);
                break;
            case GPIB_SET_CONFIG:
                gp[unit].gpib_addr = config->address;
                gp_write_register(unit, ADR, gp[unit].gpib_addr, 
                                  "set GPIB address");
                gp[unit].poll_timeout = config->poll_timeout;
                gp[unit].irq_timeout = config->irq_timeout;
                gp[unit].wait_timeout = config->wait_timeout;
                gp[unit].eos = config->eos;
                gp_write_register(unit, EOSR, gp[unit].eos, "set EOS byte");
                gp[unit].eoi_enabled = config->eoi_enabled;
                gp[unit].eos_enabled = config->eos_enabled;
                return(0);
                break;
            default:  /* otherwise, continue on */
        }

        /* all the others require the bus */
    if(!gp[unit].controller)
        return EPERM;
    if(!gp_wait_for_bus(unit, gpaddr))
        return EIO;
    if(gpaddr == GPIB_STTY_MSG)  /* device is special "global" controller */
    {
            /* these functions are global to the whole bus */
        switch (cmd) 
        {
            case GPIB_TRIGGER:
                error = gp_controller_cmd(unit, GPIB_GET);
                break;
            case GPIB_DEVICE_CLEAR:
                error = gp_DCL(unit);
                break;
            case GPIB_INTERFACE_CLEAR:
                error = gp_IFC(unit);
                break;
        }
    }
    else
    {
            /* these functions operate on one device only */
        switch (cmd)
        {
            case GPIB_LOCAL:
                error = gp_GTL(unit, gpaddr);
                break;
            case GPIB_LOCKOUT:
                error = gp_LLO(unit, gpaddr);
                break;
            case GPIB_SELECTED_DEVICE_CLEAR:
                error = gp_SDC(unit, gpaddr);
                break;
            case GPIB_SPOLL:
                error = gp_spoll(unit, gpaddr, spoll_result);
                break;
            case GPIB_REMOTE_ENABLE:
                dbg_print("gp[%d]: Remote Enable\n", unit);
                gp_write_register(unit, AUXMR, AUX_SREN, "set REN");
		error = gp_listen(unit, gpaddr);
                break;
            case GPIB_REMOTE_DISABLE:
                dbg_print("gp[%d]: Remote Disable\n", unit);
                gp_write_register(unit, AUXMR, AUX_CREN, "clear REN");
                break;
            default:
                error = ENOTTY;
        }
    }
    
    gp_release_bus(unit, gpaddr);
    return(error);
}

void gpintr(int unit)
{
    /* make sure the interrupt was for me */
    if(!(gp_read_register_quiet(unit, ISR3) & ISR3_INT))
        return;

    /* clear interrupt */
    gp_write_register_quiet(unit, IMR3, 0x00);
    wakeup(GP_INT(unit));
}

int gp_controller_cmd(int unit, unsigned char cmd)
{
    int success;

    gp_write_register_quiet(unit, IMR2, ISR2_CO);
    gp_write_register_quiet(unit, IMR1, ISR1_ERR);

#ifdef GPIB_PARANOID
    if(!gp_wait(unit, ISR3_TCLINT, "cmd", "cdor")) return EIO;
#endif

    gp_write_register_quiet(unit, CDOR, cmd);

#ifdef GPIB_DEBUG
    success = gp_wait(unit, ISR3_TCLINT, "cmd", gpib_commands[cmd]);
#else
    success = gp_wait(unit, ISR3_TCLINT, "cmd", (char*)NULL);
#endif
    if(gp_read_register_quiet(unit, ISR1) & ISR1_ERR)
    {
        dbg_print("gp[%d]: no listeners\n", unit);
	gp_error = GPIB_NOLISTEN;
        return EIO;
    }
    if(!success/*  || !(gp_read_isr2(unit) & ISR2_CO) */)
    {
	gp_error = GPIB_TIMEOUT;
/* 	gp_write_register(unit, AUXMR, 0x0E, "nbaf"); */
        return EIO;
    }
    gp_error = GPIB_NOERROR;
    return 0;
}

/* reenable controller */
int gp_controller_active(int unit)
{
        /* check to see if I already am already controller active */
    if((gp_read_adsr(unit) & 0xC0) == 0x80)
	goto done;

    gp_write_register(unit, AUXMR, AUX_TCA, "take control");

        /* wait for CDOR to be empty before going to active */
    gp_write_register_quiet(unit, IMR2, ISR2_CO);
    if(!gp_wait(unit, ISR3_TCLINT, "cmd", "controller active"))
        return EIO;

#ifdef GPIB_PARANOID
    if((gp_read_adsr(unit) & 0xC0) != 0x80) /* unsuccessful operation */
    {
        dbg_print("gp[%d]: failed to take control\n", unit);
        return 1;
    }
#endif

 done:
    gp_print_controller_state(unit);
    return 0;
}

int gp_controller_standby(int unit)
{
        /* check to see if I already am already controller standby */
    if((gp_read_adsr(unit) & 0xC0) == 0xC0)
        goto done;

    gp_write_register(unit, AUXMR, AUX_GTS, "controller to standby");
#ifdef GPIB_PARANOID
    if((gp_read_adsr(unit) & 0xC0) != 0xC0) /* unsuccessful operation */
    {
        dbg_print("gp[%d]: failed to go to standby\n", unit);
        return 1;
    }
#endif
 done:
    gp_print_controller_state(unit);
    return 0;
}

int gp_controller_idle(int unit)
{
        /* check to see if I already am already controller idle */
    if(gp_read_adsr(unit) & 0xC0 == 0x40)
/*     if((gp_read_adsr(unit) & 0xC0) == 0x40) */
    {
        gp_print_controller_state(unit);
        return 0;
    }

        /* reset bus and relinquish control */
    gp_IFC(unit);

    gp_write_register(unit, CMDR, CMDR_SOFT_RESET, "soft reset");
    gp_write_register(unit, CMDR, CMDR_CLRSC, "disable system controller");
    gp_write_register(unit, AUXMR, AUX_RST, "7210 chip reset");
    return 0;
}

int gp_controller_in_charge(int unit)
{
    if(gp_read_register_quiet(unit, ADSR) & 0x80)
    {
        dbg_print("gp[%d]: is CIC\n", unit);
	return 1;
    }
    else
    {
        dbg_print("gp[%d]: is not CIC\n", unit);
	return 0;
    }
}

int gp_become_controller_in_charge(int unit)
{
    return(gp_controller_in_charge(unit));
}

/* gp_listen() and gp_talk() require that the device be opened as a
   controller, and that the controller is active (call
   gp_controller_active() before this if you aren't sure).  These
   prerequisites are not checked here because there are somewhat
   expensive to check (ISA reads), and usually you already know. */
int gp_listen(int unit, int gpaddr)
{
    if(gp[unit].listener != gpaddr)
    {
        if(gp_UNL(unit)) return 1;

	/* Why doesn't this work? */
/*         gp_write_register(unit, AUXMR, AUX_SREN, "set REN"); */
	gp_write_register(unit, BSR, 1, "assert REN");

        if(gp_controller_cmd(unit, GPIB_MLA | gpaddr)) return 1;
        gp[unit].listener = gpaddr;
    }        
    return 0;
}

int gp_talk(int unit, int gpaddr)
{
        /* some devices have to be re-addressed to talk to get any
           data after the first message.  Therefore, we can't skip
           this step like in gp_listen().  It doesn't appear to slow
	   down most operations, anyway. */

    if(gp_UNT(unit)) return 1;
    if(gp_controller_cmd(unit, GPIB_MTA | gpaddr)) return 1;
    gp[unit].talker = gpaddr;
    return 0;
}

int gp_IFC(int unit)
{
    dbg_print("gp[%d]: Interface Clear\n", unit);
    gp_write_register(unit, AUXMR, AUX_SIFC, "set IFC");
    DELAY(100); /* 100 us */
    gp_write_register(unit, AUXMR, AUX_CIFC, "clear IFC");
    return 0;
}

int gp_DCL(int unit)
{
    if(gp_controller_cmd(unit, GPIB_DCL)) return EIO;
    return 0;
}

int gp_LLO(int unit, int address)
{
    if(gp_listen(unit, address)) return EIO;
    if(gp_controller_cmd(unit, GPIB_LLO)) return EIO;
    if(gp_UNL(unit)) return EIO;
    return 0;
}

int gp_SDC(int unit, int address)
{
    if(gp_listen(unit, address)) return EIO;
    if(gp_controller_cmd(unit, GPIB_SDC)) return EIO;
    if(gp_UNL(unit)) return EIO;
    return 0;
}

int gp_GTL(int unit, int address)
{
    if(gp_listen(unit, address)) return EIO;
    if(gp_controller_cmd(unit, GPIB_GTL)) return EIO;
    if(gp_UNL(unit)) return EIO;
    return 0;
}

int gp_UNL(int unit)
{
    if(gp_controller_cmd(unit, GPIB_UNL)) return EIO;
    gp[unit].listener = -1;
    return 0;
}

int gp_UNT(int unit)
{
    if(gp_controller_cmd(unit, GPIB_UNT)) return EIO;
    gp[unit].talker = -1;
    return 0;
}

void gp_write_count32(int unit, int count)
{
    int iobase = gp[unit].iobase;
    gp_write_register_quiet(unit, CNT0, count);
    gp_write_register_quiet(unit, CNT1, count>>8);
    gp_write_register_quiet(unit, CNT2, count>>16);
    gp_write_register_quiet(unit, CNT3, count>>24);
    dbg_print("gp[%d]: writing count %d\n", unit, count);
}

void gp_write_count(int unit, int count)
{
    int iobase = gp[unit].iobase;
    count = count;
    gp_write_register_quiet(unit, CNT0, count);
    gp_write_register_quiet(unit, CNT1, count>>8);
    dbg_print("gp[%d]: writing count %d\n", unit, count);
}

int gp_read_count32(int unit)
{
    int iobase = gp[unit].iobase;
    int count = 
        (gp_read_register_quiet(unit, CNT0) | 
         (gp_read_register_quiet(unit, CNT1) << 8) |
         (gp_read_register_quiet(unit, CNT2) << 16) |
         (gp_read_register_quiet(unit, CNT3) << 24));
    dbg_print("gp[%d]: FIFO count = %d\n", unit, count);
    return count;
}

int gp_read_count(int unit)
{
    int iobase = gp[unit].iobase;
    int count = 
        (gp_read_register_quiet(unit, CNT0) | 
         (gp_read_register_quiet(unit, CNT1) << 8));
    if(count&0x8000)
        count |= 0xFFFF0000;
    dbg_print("gp[%d]: FIFO count = %d\n", unit, count);
    return count;
}

int gp_spoll(int unit, int gpaddr, unsigned char* result)
{
    int success, time, irq_timeout, poll_timeout;

    success = 0;
    dbg_print("gp[%d]: serial poll device %d\n", unit, gpaddr);
        /* address controller to listen, device to talk */
    if(gp_listen(unit, gp[unit].gpib_addr)) goto spdone;
    if(gp_talk(unit, gpaddr)) goto spdone;

        /* serial poll */
    if(gp_controller_cmd(unit, GPIB_SPE)) goto spdone;
    if(gp_controller_standby(unit)) goto spdone;
        
        /* read status byte.  Change the timeouts, because there will
	   often be no talker and the wait() will timeout.  We don't
	   want this to take too long. */
    irq_timeout = gp[unit].irq_timeout;
    poll_timeout = gp[unit].poll_timeout;
    gp[unit].poll_timeout = 10;
    gp[unit].irq_timeout = 10;
    gp_write_register(unit, IMR1, ISR1_DI, "enable interrupts");
    success = gp_wait(unit, ISR3_TCLINT, "spoll", "spoll");
    if(success)
        *result = gp_read_register(unit, DIR);
    gp_write_register(unit, IMR1, 0x00, "disable ints");
    gp[unit].poll_timeout = poll_timeout;
    gp[unit].irq_timeout = irq_timeout;
    
        /* disable serial poll */
 spdone:
    gp_controller_active(unit);
    gp_controller_cmd(unit, GPIB_SPD);
    if(success)
        return 0;
    else
        return EIO;
}

/* the basic idea of this probe is stolen from the Linux driver. */
int gp_atgpib_probe(int unit)
{
    int i;

    dbg_print("gp[%d]: checking for GPIB-AT\n", unit);

    gp_write_register(unit, CMDR, CMDR_SOFT_RESET, "soft reset");
    gp_write_register(unit, CMDR, CMDR_CLRSC, "disable system controller");
    gp_write_register(unit, AUXMR, 0x02, "reset");
    gp_write_register(unit, AUXMR, AUX_RST, "7210 chip reset");
        /* Check if FIFO accepts bytes */
    gp_write_register_quiet(unit, CFG, 0x00 /*, "" */);
    gp_write_register_quiet(unit, CMDR, 0x10 /*, "reset FIFO" */);
    gp_write_register_quiet(unit, CNT0, -16 /*, "byte count low register" */);
    gp_write_register_quiet(unit, CNT1, -16 >> 8 /*, "byte count high register" */);
  
    if (!(gp_read_register_quiet(unit, ISR3) & (ISR3_NFF | ISR3_NEF)))
    {
        dbg_print("gp[%d]: no NFF or NEF - probe failed!\n", unit);
        return 0;
    }   /* FIFO not empty or full or not NI board*/
  
    for(i = 0; i < 20; i++) 
    {
        gp_write_register_quiet(unit, FIFOB, i*3/* , "junk data" */);
        if (!(gp_read_register_quiet(unit, ISR3) & ISR3_NFF))
            break;
    }
        /* FIFO should be full exactly after 16 byte writes */
    if (i != 15)
    {
        dbg_print("gp[%d]: FIFO didn't fill up on probe.\n", unit);
        return 0;
    }

        /* Now read bytes back from FIFO */
    gp_write_register_quiet(unit, CNT0, -16/* , "byte count low register" */);
    gp_write_register_quiet(unit, CNT1, -16 >> 8/* , "byte count high register" */);
    gp_write_register_quiet(unit, CFG, 0x28/* , "IN and CCEN" */);
    gp_write_register_quiet(unit, CMDR, 0x04/* , "go" */);

    for(i = 0; i < 20; i++) 
    {
        if (gp_read_register_quiet(unit, FIFOB) != i*3)
	{
            dbg_print("gp[%d]: FIFO read check failed\n", unit);
            return 0;		
	}
        if (!(gp_read_register_quiet(unit, ISR3) & ISR3_NEF))
            break;
    }
        /* FIFO should be empty exactly after 16 byte reads */
    if (i != 15)
        return 0;

    return 1;
}

void gp_atgpib_attach(int unit)
{
    dbg_print("gp: attaching AT-GPIB at 0x%x\n", gp[unit].iobase);

    gp_write_register(unit, CMDR, CMDR_SOFT_RESET, "soft reset");
    gp_write_register(unit, CMDR, CMDR_CLRSC, "disable system controller");
    gp_write_register(unit, AUXMR, 0x02, "reset");
    gp_write_register(unit, AUXMR, AUX_RST, "7210 chip reset");

        /* clear status registers */
    (void)gp_read_register(unit, ISR0);
    (void)gp_read_register(unit, ISR1);
    (void)gp_read_register(unit, ISR2);
    (void)gp_read_register(unit, ISR3);

    gp_write_register(unit, IMR0, 0x00, "disable all interrupts");
    gp_write_register(unit, IMR1, 0x00, "disable all interrupts");
    gp_write_register(unit, IMR2, 0x00, "disable all interrupts");
    gp_write_register(unit, IMR3, 0x00, "disable all interrupts");

        /* configure addressing */
    gp_write_register(unit, ADMR, 0x31, "dual primary addressing mode");
    gp_write_register(unit, ADR, gp[unit].gpib_addr, "set gpib address");
    gp_write_register(unit, ADR, 0xE0, "disable extended addresses");

    gp_write_register(unit, EOSR, gp[unit].eos, "set end of string register");
    gp_write_register(unit, AUXMR, 0xA0, "AUXRB - active high interrupts");
    gp_write_register(unit, HSSEL, 0x00, "enable DMA");
    if(gp[unit].irq)
    {
        gp_write_register(unit, IMR0, 0x00, "no interrupts");
        gp_write_register(unit, IMR1, 0x00, "no interrupts");
        gp_write_register(unit, IMR2, 0x00, "no interrupts");
        gp_write_register(unit, IMR3, 0x00, "no interrupts");
        gp_write_register(unit, INTR, 0x01, "enable ISA interrupts");
    }
    else
        gp_write_register(unit, INTR, 0x00, "disable ISA interrupts");
}

int gp_programmed_read(int unit, struct uio* uio)
{
    int success, end, count, tsize, i;

    dbg_print("gp[%d]: performing programmed read\n", unit);

    gp_write_register(unit, AUXMR, 0x82, "AUXRA, RFD holdoff on END");
    gp_write_register(unit, CFG, 0x77, "16 bits from FIFOA");
    gp_write_count(unit, 0);
    gp_write_register(unit, IMR1, ISR1_END, "interrupt on END");
    gp_write_register(unit, IMR2, 0x10, "enable DMAI");
    gp_write_register(unit, CMDR, CMDR_GO, "go");

    end = 0;
    while((!end || (gp_read_isr3(unit)&ISR3_NEF)) && (uio->uio_resid > 0))
    {
        dbg_print("gp[%d]: reading next uio chunk\n", unit);
        tsize = min(GPIB_MAX_TRANSFER, uio->uio_resid);
        count = 0;
        while((!end || (gp_read_isr3(unit)&ISR3_NEF)) && (count < tsize))
        {
            success = gp_wait(unit, ISR3_NEF | ISR3_DONE, "read", "NEF");
            if(!success)
            {
                gp_error = GPIB_TIMEOUT;
		success = 0;
		goto done;
            }
            while((gp_read_isr3(unit) & ISR3_NEF) && (count < tsize))
            {
	        unsigned int w = gp_read_register_w(unit, FIFOB);
		gp[unit].buf[count + 1] = w&0xFF;
                gp[unit].buf[count] = (w&0xFF00)>>8;
                count += 2;
            }
            if(!end && (gp_read_isr1(unit) & ISR1_END))
	    {
	        gp_write_register(unit, CMDR, CMDR_STOP, "stop");
                end = 1;
	    }
        }
	if(end && !(gp_read_isr3(unit)&ISR3_NEF))
	{
	    /* forget last garbage byte, if the transfer size was odd */ 
	    if(gp_read_count(unit) & 0x01)
	        count--;
	
	    /* erase terminator, if any */
	    if(gp[unit].eos_enabled)
  	        if(gp[unit].buf[count-1] == gp[unit].eos)
		    count--;
	    gp[unit].buf[count] = 0;
	}
        if(uiomove(gp[unit].buf, count, uio))
        {
            dbg_print("gp[%d]: uiomove failed in read\n", unit);
	    success = 0;
	    goto done;
        }
    }
    if(end)
    {
        dbg_print("gp[%d]: END detected\n", unit);
	gp_write_register(unit, CMDR, CMDR_RESET_FIFO, "reset FIFO");
        gp_write_register(unit, AUXMR, AUX_RHDF, "clear END holdoff");
        success = 1;
    }
    else
    {
        dbg_print("gp[%d]: buffer overflow\n", unit);
        gp_error = GPIB_OVERFLOW;
        success = 0;
    }

 done:
    gp_write_register(unit, CMDR, CMDR_STOP, "stop");
    gp_write_register(unit, IMR1, 0x00, "disable ints");
    gp_write_register(unit, IMR2, 0x00, "disable DMAI");
    return success;
}

int gp_dma_read(int unit, struct uio* uio)
{
    int success, tsize, dmasize, end, count, i;
    dbg_print("gp[%d]: performing read with DMA\n", unit);

    gp_write_register(unit, AUXMR, 0x86, "AUXRA, RFD holdoff on END");
    gp_write_register(unit, CFG, 0x65, "16 bits to FIFOB");
    gp_write_register(unit, IMR1, ISR1_END, "interrupt on END");
    gp_write_register(unit, IMR2, 0x10, "enable DMA");
    gp_write_register(unit, TIMER, 0x88, "timout 12us");
    gp_write_register(unit, ACCWR, 0x01, "enable DMA");

    end = 0;
    success = 1;
    while(success && uio->uio_resid && 
	  (!end/*  || (gp_read_isr3(unit) & ISR3_NEF) */))
    {
        dbg_print("gp[%d]: reading next uio chunk\n", unit);
        dmasize = tsize = min(GPIB_MAX_TRANSFER, uio->uio_resid);
	gp_write_count(unit, -tsize);
	gp_write_register(unit, CMDR, CMDR_GO, "go");
        if(dmasize & 0x01) /* if dmasize is odd, make it even */
            dmasize++;
        isa_dmastart(B_READ, gp[unit].buf, dmasize, gp[unit].drq);
	if(!end)
	{
	    success = gp_wait(unit, ISR3_DONE | ISR3_TCLINT, "read", "DONE");
	    if(!end && (gp_read_isr1(unit) & ISR1_END))
	        end = 1;
	}
	dbg_print("gp[%d]: dma transferred %d bytes\n", unit, 
		  dmasize - isa_dmastatus(gp[unit].drq));
        isa_dmastop(gp[unit].drq);
	isa_dmadone(B_READ, gp[unit].buf, dmasize, gp[unit].drq);
	gp_write_register(unit, CMDR, CMDR_STOP, "stop");
	count = tsize + gp_read_count(unit);
	dbg_print("gp[%d]: actual transfer count: %d\n", unit, count);
        if(!success)
        {
            dbg_print("gp[%d]: read timeout\n", unit);
            gp_error = GPIB_TIMEOUT;
        }
	if(end)
	{
	    /* erase terminator, if any */
	    if(gp[unit].eos_enabled && 
	       (gp[unit].buf[count-1] == gp[unit].eos))
	        count--;
	    gp[unit].buf[count] = 0;
	}
        if(uiomove(gp[unit].buf, count, uio))
        {
            dbg_print("gp[%d]: uiomove failed in read\n", unit);
            success = 0;
        }
    }
    gp_write_register(unit, ACCWR, 0x00, "disable DMA");
    gp_write_register(unit, CMDR, CMDR_RESET_FIFO, "reset FIFO");
    if(success)
    {
        dbg_print("gp[%d]: got END\n", unit);
	gp_write_register(unit, AUXMR, AUX_RHDF, "clear END holdoff");
    }

    return success;
}

/* return 1 on success */
int gp_programmed_send(int unit, struct uio* uio)
{
    int count, success, tsize;
    dbg_print("gp[%d]: performing programmed send\n", unit);
 
       /* configure terminators */
    if(gp[unit].eoi_enabled)
    {
        gp_write_register(unit, CCR, AUX_SEOI, "send EOI on carry");
	gp_write_register(unit, CFG, 0x5F, "16 bit transfers from FIFOA, CCEN");
    }
    else
	gp_write_register(unit, CFG, 0x57, "16 bit transfers from FIFOA, no CCEN");
    
    gp_write_register(unit, CMDR, CMDR_RESET_FIFO, "reset FIFO");
    if(gp[unit].eos_enabled)
        gp_write_count(unit, -(uio->uio_resid + 1));
    else
        gp_write_count(unit, -uio->uio_resid);
    gp_write_register(unit, IMR1, ISR1_ERR, "enable ERR interrupt");
    gp_write_register(unit, IMR2, 0x20, "enable DMA");
    gp_write_register(unit, CMDR, CMDR_GO, "go");

    while(uio->uio_resid > 0)
    {
        dbg_print("gp[%d]: sending next uio chunk\n", unit);
        tsize = min(GPIB_MAX_TRANSFER, uio->uio_resid);
        if(uiomove(gp[unit].buf, tsize, uio))
        {
            dbg_print("gp[%d]: uiomove failed in send\n", unit);
	    success = 0;
	    goto done;
        }
        count = 0;

	/* tack EOS onto end of message, if required */
	/* FIXME: this is a array bounds bug if tsize == GPIB_MAX_TRANSFER 
	   and it is also the last uio chunk */
	if((uio->uio_resid == 0) && gp[unit].eos_enabled)
/* 	  && (gp[unit].buf[tsize-1] != gp[unit].eos)) */
	{
	    gp[unit].buf[tsize] = gp[unit].eos;
	    tsize++;
	}
        if(tsize & 0x01)  /* odd number of bytes, make it even */
            tsize++;

        while(count < tsize)
        {
            if(!(gp_read_isr3(unit) & ISR3_NFF))
            {
                if(gp_read_isr1(unit) & ISR1_ERR)
                {
                    dbg_print("gp[%d]: no listeners\n", unit);
                    gp_error = GPIB_NOLISTEN;
		    success = 0;
		    goto done;
                }
                    /* wait for space in FIFO */
                success = 
                    gp_wait(unit, ISR3_NFF | ISR3_TCLINT | ISR3_DONE, 
                                          "write", "NFF");
                if(!success)
                {
                    dbg_print("gp[%d]: write timeout\n", unit);
		    success = 0;
                    gp_error = GPIB_TIMEOUT;
		    goto done;
                }
                if(gp_read_isr3(unit) & (ISR3_TCLINT | ISR3_DONE))
                { 
                    dbg_print("gp[%d]: premature DONE!\n", unit);
		    success = 0;
                    gp_error = GPIB_UNKNOWN;
		    goto done;
                }
                    /* otherwise, we're good to go... */
            }
            gp_write_register_w(unit, FIFOB, 
                                gp[unit].buf[count+1] | 
                                (gp[unit].buf[count]<<8), "write");
            count += 2;
        }
    }
    success = gp_wait(unit, ISR3_DONE, "write", "DONE");
 done:
    gp_write_register(unit, CMDR, CMDR_STOP, "stop");
    gp_write_register(unit, IMR1, 0x00, "disable ints");
    gp_write_register(unit, IMR2, 0x00, "disable DMA");
    return success;
}

int gp_dma_send(int unit, struct uio* uio)
{
    int success, tsize;
    dbg_print("gp[%d]: performing send with DMA\n", unit);

    gp_write_register(unit, AUXMR, 0x82, "AUXRA");
    gp_write_register(unit, CFG, 0x45, "16 bit transfers from FIFOA, no CCEN");
    gp_write_register(unit, CMDR, CMDR_RESET_FIFO, "reset FIFO");
    gp_write_register(unit, IMR1, ISR1_ERR, "enable ERR interrupt");
    gp_write_register(unit, IMR2, 0x20, "enable DMA");
    gp_write_register(unit, TIMER, 0xE2, "timout 30 transfers");
    gp_write_register(unit, ACCWR, 0x01, "enable DMA");

    success = 1;
    while(success && (uio->uio_resid > 0))
    {
        tsize = min(GPIB_MAX_TRANSFER, uio->uio_resid);
        if(uiomove(gp[unit].buf, tsize, uio))
        {
            dbg_print("gp[%d]: uiomove failed in send\n", unit);
            success = 0;
            goto done;
        }
	/* tack EOS onto end of message, if required */
	/* FIXME: this is a array bounds bug if tsize == GPIB_MAX_TRANSFER 
	   and it is also the last uio chunk */
	if((uio->uio_resid == 0) && gp[unit].eos_enabled
	   && (gp[unit].buf[tsize-1] != gp[unit].eos))
	{
	    gp[unit].buf[tsize] = gp[unit].eos;
	    tsize++;
	}
        dbg_print("gp[%d]: sending uio chunk (%d bytes)\n", unit, tsize);
        gp_write_count(unit, -tsize);
	/* configure terminators on last uio chunk*/
	if((uio->uio_resid == 0) && gp[unit].eoi_enabled)
	{
	    gp_write_register(unit, CCR, AUX_SEOI, "send EOI on carry");
	    gp_write_register(unit, CFG, 0x4D, "16 bits from FIFOA, CCEN");
	}
        gp_write_register(unit, CMDR, CMDR_GO, "go");
        if(tsize&0x1) /* if tsize is odd, make it even.  This should
                         only happen on the last uio chunk, if at all. */
            tsize++;
        isa_dmastart(B_WRITE, gp[unit].buf, tsize, gp[unit].drq);
	if(uio->uio_resid == 0)
	    success = gp_wait(unit, ISR3_DONE|ISR3_TCLINT, "write", "DONE");
	else
	    success = gp_wait(unit, ISR3_STOP|ISR3_TCLINT, "write", "STOP");
        gp_write_register(unit, CMDR, CMDR_STOP, "stop");
	isa_dmastop(gp[unit].drq);
	isa_dmadone(B_WRITE, gp[unit].buf, tsize, gp[unit].drq);
	dbg_print("gp[%d]: dma transferred %d bytes\n", unit, 
		  tsize - isa_dmastatus(gp[unit].drq));
	if(!success)
        {
            dbg_print("gp[%d]: write timeout\n", unit);
            gp_error = GPIB_TIMEOUT;
	    goto done;
        }
        else if(gp_read_isr1(unit) & ISR1_ERR)
        {
            dbg_print("gp[%d]: no listeners\n", unit);
            gp_error = GPIB_NOLISTEN;
            success = 0;
	    goto done;
        }
    }

done:
    gp_write_register(unit, ACCWR, 0x00, "disable DMA");
    return success;
}

void gp_print_spoll(int unit, char* result)
{
    int paddr;
    char msg[256];
    unsigned char sp;
    int resultPtr = 0;

    if(gp[unit].spollmessage > 31)
    {
        result[0] = 0;
	gp[unit].spollmessage = 0;
	return;
    }
    if(!gp_wait_for_bus(unit, GPIB_POLL_PORT))
    {
        sprintf(result, "\t\t%d: cannot get bus\n", 
		gp[unit].spollmessage);
	return;
    }

    if(gp_spoll(unit, gp[unit].spollmessage, &sp) == 0)
        sprintf(result, "\t\t%d: 0x%x\n", gp[unit].spollmessage, sp);
    else
        sprintf(result, "\t\t%d: \n", gp[unit].spollmessage);

    gp_release_bus(unit, 31);
    gp[unit].spollmessage++;
    return;
}

void gp_print_stty(int unit, char* result)
{
    char msg[256];
    int resultPtr = 0;

    if(gp[unit].sttymessage)
    {
	result[0] = 0;
        gp[unit].sttymessage = 0;
	return;
    }
    gp[unit].sttymessage = 1;
    switch(gp[unit].type)
    {
        case TNT4882:
            sprintf(msg, "gp[%d]: AT-GPIB/TNT at 0x%x", unit,
                    gp[unit].iobase);
            break;
        case NAT4882A:
        case NAT4882B:
        case TURBO:
            sprintf(msg, "gp[%d]: AT-GPIB at 0x%x", unit,
                    gp[unit].iobase);
            break;
        }
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    if(gp[unit].irq)
        sprintf(msg, "\t, IRQ %d", ffs(gp[unit].irq)-1);
    else
        sprintf(msg, "\t using polled I/O");
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    if(gp[unit].drq_aquired)
    {
        sprintf(msg, ", on DMA channel %d", gp[unit].drq);
        memcpy(&result[resultPtr], msg, strlen(msg));
        resultPtr += strlen(msg);
    }
    sprintf(msg, "\n");
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    if(gp_read_adsr(unit) & 0x80)
        sprintf(msg, "\tCIC (controller in charge)\n"); 
    else
        sprintf(msg, "\tnot CIC\n"); 
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    sprintf(msg, "\tpoll timeout: %d\n", gp[unit].poll_timeout);
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    if(gp[unit].wait_timeout > 0)
        sprintf(msg, "\tinterrupt timeout: %d\n",
		gp[unit].irq_timeout);
    else
        sprintf(msg, "\tirq timeout: none\n");
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    if(gp[unit].wait_timeout > 0)
        sprintf(msg, "\tbus wait timeout: %d\n",
		gp[unit].wait_timeout);
    else
        sprintf(msg, "\tbus wait timeout: none\n");
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    if(gp[unit].eoi_enabled)
        sprintf(msg, "\tEOI enabled\n");
    else
        sprintf(msg, "\tEOI disabled\n");
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    if(gp[unit].eos_enabled)
        sprintf(msg, "\tEOS enabled, using 0x%x\n", gp[unit].eos);
    else
        sprintf(msg, "\tEOS disabled\n");
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    sprintf(msg, "\tcounter size (maxtransfer): %d\n",
            gp[unit].maxtransfer);
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    return;
}

void gp_print_status(int unit, char* result)
{
    char msg[32];
    int resultPtr = 0;
    unsigned char adsr;

    if(gp[unit].statusmessage)
    {
	result[0] = 0;
        gp[unit].statusmessage = 0;
	return;
    }
    gp[unit].statusmessage = 1;

    adsr = gp_read_register_quiet(unit, ADSR);
    if(adsr & 0x80)
        sprintf(msg, "+CIC"); 
    else
        sprintf(msg, "-CIC"); 
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    if(adsr & 0x40)
        sprintf(msg, " +ATN"); 
    else
        sprintf(msg, " -ATN"); 
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    if(adsr & 0x20)
        sprintf(msg, " +SPMS"); 
    else
        sprintf(msg, " -SPMS"); 
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    if(adsr & 0x10)
        sprintf(msg, " +LPAS"); 
    else
        sprintf(msg, " -LPAS"); 
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    if(adsr & 0x08)
        sprintf(msg, " +TPAS"); 
    else
        sprintf(msg, " -TPAS"); 
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    if(adsr & 0x04)
        sprintf(msg, " +LA"); 
    else
        sprintf(msg, " -LA"); 
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    if(adsr & 0x02)
        sprintf(msg, " +TA"); 
    else
        sprintf(msg, " -TA"); 
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
    if(adsr & 0x01)
        sprintf(msg, " +MJMN\n"); 
    else
        sprintf(msg, " -MJMN\n"); 
    memcpy(&result[resultPtr], msg, strlen(msg));
    resultPtr += strlen(msg);
}

#ifdef GPIB_DEBUG

int gp_read_isr0(int unit)
{
    int isr = gp_read_register_quiet(unit, ISR0);
    dbg_print("gp[%d] isr0: ", unit);
    if(isr&0x80)
        dbg_print("nba ");
    if(isr&0x40)
        dbg_print("STBO ");
    if(isr&0x20)
        dbg_print("NL ");
    if(isr&0x10)
        dbg_print("EOS ");
    if(isr&0x08)
        dbg_print("IFCI ");
    if(isr&0x04)
        dbg_print("ATNI ");
    if(isr&0x02)
        dbg_print("TO ");
    if(isr&0x01)
        dbg_print("SYNC ");
    dbg_print("\n");
    return isr;
}

int gp_read_isr1(int unit)
{
    int isr = gp_read_register_quiet(unit, ISR1);
    dbg_print("gp[%d]: isr1 --> ", unit);
    if(isr&0x80)
        dbg_print("CPT ");
    if(isr&0x40)
        dbg_print("APT ");
    if(isr&0x20)
        dbg_print("DET ");
    if(isr&0x10)
        dbg_print("END_RX ");
    if(isr&0x08)
        dbg_print("DEC ");
    if(isr&0x04)
        dbg_print("ERR ");
    if(isr&0x02)
        dbg_print("DO ");
    if(isr&0x01)
        dbg_print("DI ");
    dbg_print("\n");
    return isr;
}

int gp_read_isr2(int unit)
{
    int isr = gp_read_register_quiet(unit, ISR2);
    dbg_print("gp[%d]: isr2 --> ", unit);
    if(isr&0x80)
        dbg_print("INT ");
    if(isr&0x40)
        dbg_print("SRQI ");
    if(isr&0x20)
        dbg_print("LOK ");
    if(isr&0x10)
        dbg_print("REM ");
    if(isr&0x08)
        dbg_print("CO ");
    if(isr&0x04)
        dbg_print("LOKC ");
    if(isr&0x02)
        dbg_print("REMC ");
    if(isr&0x01)
        dbg_print("ADSC ");
    dbg_print("\n");
    return isr;
}

int gp_read_isr3(int unit)
{
    int isr = gp_read_register_quiet(unit, ISR3);
    dbg_print("gp[%d]: isr3 --> ", unit);
    if(isr&0x80)
        dbg_print("INT ");
    if(isr&0x40)
        dbg_print("INTSRC2 ");
    if(isr&0x10)
        dbg_print("STOP ");
    if(isr&0x08)
        dbg_print("NFF ");
    if(isr&0x04)
        dbg_print("NEF ");
    if(isr&0x02)
        dbg_print("TLCINT ");
    if(isr&0x01)
        dbg_print("DONE ");
    dbg_print("\n");
    return isr;
}

int gp_read_register(int unit, int reg)
{
    int result;
    int ioaddr = gp[unit].iobase + reg;

    result = inb(ioaddr);
    dbg_print("gp[%d]: read 0x%x from 0x%x(%s)\n", unit, result, ioaddr,
              gpib_read_registers[reg]);
    return result;
}

int gp_read_register_w(int unit, int reg)
{
    int result;
    int ioaddr = gp[unit].iobase + reg;

    result = inw(ioaddr);
    dbg_print("gp[%d]: read 0x%x from 0x%x(%s)\n", unit, result, ioaddr,
              gpib_read_registers[reg]);
    return result;
}

void gp_write_register(int unit, int reg, int val, const char* msg)
{
    int ioaddr = gp[unit].iobase + reg;
    outb(ioaddr, val&0xFF);
    dbg_print("gp[%d]: write 0x%x to 0x%x.(%s) -- %s\n", unit, val&0xFF, 
	      ioaddr, gpib_write_registers[reg], msg);
}

void gp_write_register_quiet(int unit, int reg, int val)
{
    outb(gp[unit].iobase + reg, val&0xFF);
}

void gp_write_register_w(int unit, int reg, int val, const char* msg)
{
    int ioaddr = gp[unit].iobase + reg;
    outw(ioaddr, val);
    dbg_print("gp[%d]: write 0x%x to 0x%x.(%s) -- %s\n", unit, val, ioaddr,
              gpib_write_registers[reg], msg);
}

void gp_print_bus(int unit)
{
    int bus;
    if(gp[unit].type == TURBO)
        return;  /* my at-gpib can't read the bus, apparently */
    bus = gp_read_register(unit, BSR);
    dbg_print("gp[%d]: bus --> ", unit);
    if(bus&0x80)
        dbg_print("ATN ");
    if(bus&0x40)
        dbg_print("DAV ");
    if(bus&0x20)
        dbg_print("NDAC ");
    if(bus&0x10)
        dbg_print("NRFD ");
    if(bus&0x08)
        dbg_print("EO1 ");
    if(bus&0x04)
        dbg_print("SRQ ");
    if(bus&0x02)
        dbg_print("IFC ");
    if(bus&0x01)
        dbg_print("REN ");
    dbg_print("\n");
}

int gp_read_adsr(int unit)
{
    int adsr = gp_read_register_quiet(unit, ADSR);
    dbg_print("gp[%d]: adsr --> ", unit);
    if(adsr&0x80)
        dbg_print("CIC ");
    if(!(adsr&0x40))
        dbg_print("*ATN ");
    if(adsr&0x20)
        dbg_print("SPMS ");
    if(adsr&0x10)
        dbg_print("LPAS ");
    if(adsr&0x08)
        dbg_print("TPAS ");
    if(adsr&0x04)
        dbg_print("LA ");
    if(adsr&0x02)
        dbg_print("TA ");
    if(adsr&0x01)
        dbg_print("MJMN ");
    dbg_print("\n");
    return adsr;
}

int gp_read_sts1(int unit)
{
    int sts = gp_read_register_quiet(unit, STS1);
    dbg_print("gp[%d]: sts1 --> ", unit);
    if(sts&0x80)
        dbg_print("DONE ");
    if(sts&0x20)
        dbg_print("IN ");
    if(sts&0x10)
        dbg_print("DRQ ");
    if(sts&0x08)
        dbg_print("STOP ");
    if(sts&0x04)
        dbg_print("DAV ");
    if(sts&0x02)
        dbg_print("HALT ");
    if(sts&0x01)
        dbg_print("GSYNC ");
    dbg_print("\n");
    return sts;
}

void gp_print_controller_state(int unit)
{
    int iobase = gp[unit].iobase;
    int adsr = inb(gp[unit].iobase + ADSR);
    switch(adsr&0xC0)
    {
        case 0x00:
        case 0x40:
            dbg_print("gp[%d]: Idle Controller\n", unit);
            break;
        case 0xC0:
            dbg_print("gp[%d]: Standby Controller\n", unit);
            break;
        case 0x80:
            dbg_print("gp[%d]: Active Controller\n", unit);
            break;
    }
}

#endif /* GPIB_DEBUG */

static gp_devsw_installed = 0;

static void gp_drvinit(void *unused)
{
    dev_t dev;

    if(!gp_devsw_installed ) 
    {
        dev = makedev(GP_CDEV_MAJOR,0);
        cdevsw_add(&dev,&gp_cdevsw,NULL);
        gp_devsw_installed = 1;
    }
}

SYSINIT(gpdev, SI_SUB_DRIVERS, SI_ORDER_MIDDLE + GP_CDEV_MAJOR, gp_drvinit,
        NULL)

#ifdef GP_MODULE

/* enable lkm stuff */

#include <sys/exec.h>
#include <sys/sysent.h>
#include <sys/sysproto.h>
#include <sys/lkm.h>

MOD_DEV(gp, LM_DT_CHAR, GP_CDEV_MAJOR, &gp_cdevsw);
int gp_stat(struct lkm_table* lkmtp, int cmd);

#define ADDR0 0x2C0
#define INT0 /* IRQ9 */ IRQ15
#define DMA0 5 /* 5 */
#define ADDR1 0x1C0
#define INT1 IRQ12
#define DMA1 6 /* 6 */

static struct isa_device dev[] = {
    {
        0,            /* device id */
        &gpdriver,   /* driver pointer */
        ADDR0,        /* base io address */
        INT0,         /* interrupt */
        DMA0,  /* dma channel */
        (caddr_t)0,   /* physical io memory */
        0,             /* size of io memory */
        gpintr,     /* interrupt interface */
        0,             /* unit number */
        0,             /* flags */
        0,             /* scsi id */
        0,             /* is alive (present) */
        0,             /* flags for register_intr() */
        0,             /* hot eject device support */
        1,             /* is device enabled */
        0,             /* we're allowed to conflict with things */
        (struct isa_device*)NULL  /* used in isa_devlist in userconfig() */
    },
    {
        0,            /* device id */
        &gpdriver,    /* driver pointer */
        ADDR1,        /* base io address */
        INT1,         /* interrupt */
        DMA1,           /* dma channel */
        (caddr_t)0,   /* physical io memory */
        0,             /* size of io memory */
        gpintr ,     /* interrupt interface */
        1,             /* unit number */
        0,             /* flags */
        0,             /* scsi id */
        0,             /* is alive (present) */
        0,             /* flags for register_intr() */
        0,             /* hot eject device support */
        1,             /* is device enabled */
        0,             /* we're allowed to conflict with things */
        (struct isa_device*)NULL  /* used in isa_devlist in userconfig() */
    }
};


int gp_load(struct lkm_table *lkmtp, int cmd)
{
    int error, unit;

    dbg_print("gp: loading gp driver\n");

    for(unit = 0; unit < NGP; unit++)
    {
        dev[unit].id_alive = gp_probe(&dev[unit]);
	if(!dev[unit].id_alive)
	{
	    dbg_print("gp[%d]: probe failed\n", unit);
	    goto fail;
	}
	gp_attach(&dev[unit]);
	
	if(dev[unit].id_irq)
        {
	    INTRMASK(tty_imask, dev[unit].id_irq);
	    error = register_intr(ffs(dev[unit].id_irq)-1,
				  dev[unit].id_id, dev[unit].id_ri_flags,
				  dev[unit].id_intr, &tty_imask,
				  dev[unit].id_unit);
	    if(error)
            {
	        uprintf("gp[%d]: cannot aquire interrupt %d\n", unit, 
			ffs(dev[unit].id_irq)-1);
		goto fail;
	    }
	    INTREN(dev[unit].id_irq);  /* enable interrupt (see icu.h) */
	}
    }

    dbg_print("gp driver loaded\n");
    return 0;

 fail:
    /* relinquish other resources */
    while(unit >= 0)
    {
        if(dev[unit].id_irq)
        {
	    INTRDIS(dev[unit].id_irq);
	    unregister_intr(ffs(dev[unit].id_irq)-1, 
			    dev[unit].id_intr);
	    INTRUNMASK(tty_imask, dev[unit].id_irq);
	}
	unit--;
    }
    return 1;
}

int gp_unload(struct lkm_table *lkmtp, int cmd)
{
    int unit;
    dbg_print("gp: unloading gp driver\n");

    for(unit = 0; unit < NGP; unit++)
    {
        if(gp[unit].open)
	{
	    dbg_print("gp: unit %d is still open\n", unit);
	    return 1;
	}
    }

    for(unit = 0; unit < NGP; unit++)
    {
        if(gp[unit].irq)
	{
	    INTRDIS(dev[unit].id_irq);
	    unregister_intr(ffs(dev[unit].id_irq)-1, dev[unit].id_intr); 
	    INTRUNMASK(tty_imask, dev[unit].id_irq);
	}
    }
    return 0;
}

int gp_stat(struct lkm_table *lkmtp, int cmd)
{
    dbg_print("gp: in gp_stat?\n");
    return 0;
}

int gp_mod(struct lkm_table *lkmtp, int cmd, int ver)
{
    DISPATCH(lkmtp, cmd, ver, gp_load, gp_unload, gp_stat);
    return 0;
}

#endif /* GP_MODULE */

/*  support PnP cards. Note that you can't use this driver as an LKM
    in this case. */

#if NPNP > 0

static char *gppnp_probe(u_long csn, u_long vend_id);
static void gppnp_attach(u_long csn, u_long vend_id, char *name,
                         struct isa_device *dev);
static u_long ngppnp = NGP;

static struct pnp_device gppnp = {
   "gppnp",
   gppnp_probe,
   gppnp_attach,
   &ngppnp,
   &tty_imask
};
DATA_SET(pnpdevice_set, gppnp);

static char *gppnp_probe(u_long csn, u_long vend_id)
{
   struct pnp_cinfo d;

   if (bootverbose)
      printf("gp: probing vendor-ID 0x%08lx\n", vend_id);

   if (vend_id != 0x01c62339)
      return NULL;

   read_pnp_parms(&d, 0);
   if (d.enable == 0)
   {
      printf("CSN %d is disabled.\n", csn);
      return NULL;
   }

   return "AT-GPIB/TNT (PnP)";
}

static void gppnp_attach(u_long csn, u_long vend_id, char *name,
                         struct isa_device *dev)
{
   struct pnp_cinfo d;
   struct isa_device *dvp;
   
   if (dev->id_unit >= NGPTOT)
      return;
   
   if (read_pnp_parms(&d, 0) == 0) {
      printf("failed to read pnp parms\n");
      return;
   }
   
   write_pnp_parms(&d, 0);
   
   enable_pnp_card();
    
   dev->id_iobase = d.port[0];
   dev->id_irq = (1 << d.irq[0]);
   dev->id_intr = gpintr;
   dev->id_ri_flags = RI_FAST;
/* don't use DMA for now */
   dev->id_drq = -1;
   
   if (dev->id_driver == NULL) {
      dev->id_driver = &gpdriver;
      dvp = find_isadev(isa_devtab_tty, &gpdriver, 0);
      if (dvp != NULL)
         dev->id_id = dvp->id_id;
   }
   
   if ((dev->id_alive = gp_probe(dev)) != 0)
      gp_attach(dev);
   else
      printf("gp%d: probe failed\n", dev->id_unit);
}

#endif  /* NPNP > 0 */
#endif  /* NGPTOT > 0 */
